# -*- coding: utf-8 -*-

"""
    sijax.plugin.upload
    ~~~~~~~~~~~~~~~~~~~

    Provides helpers to register Upload functions,
    and the :class:`sijax.plugin.comet.UploadResponse` class
    used instead of :class:`sijax.response.BaseResponse`.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from ...helper import json
from ...exception import SijaxError
from ...response import StreamingIframeResponse

# Parameters with these names are passed to the client side (JS)
# Modifying this would mean modifying the javascript files
PARAM_FORM_ID = 'formId'
PARAM_CALLBACK = 'callback'


def _prepare_options(sijax_instance, options):
    param_response_class = sijax_instance.__class__.PARAM_RESPONSE_CLASS
    if param_response_class not in options:
        options[param_response_class] = UploadResponse

    return options


def func_name_by_form_id(form_id):
    return '%s_upload' % form_id


def register_upload_callback(sijax_instance, form_id, callback, **options):
    """Helps you easily register Upload functions with Sijax.
    
    We recommend the ``args_extra`` option to be used here, so that the
    response function would receive the files object or
    anything else it might be needing to manipulate file uploads.
    By using ``args_extra`` you could change the response function's
    default signature from::

        def upload_handler(obj_response, form_values)

    to::

        def upload_handler(obj_response, files, form_values)

    where the args_extra=[files] option was passed during registering.
    To learn more on how the ``args_extra`` option works,
    refer to :meth:`sijax.Sijax.register_callback`.

    It's very important to keep in mind that unlike the regular
    :meth:`sijax.Sijax.register_callback` function, this one
    returns some javascript code as a response.
    You need to execute that code on the page that contains the
    form you want to transform to Sijax upload.

    :param sijax_instance: the :class:`sijax.Sijax` instance
                           to register callbacks with
    :param form_id: the id of the form as it appears in the DOM
    :param callback: the function to call to process the upload request
    :param options: options to pass to :meth:`sijax.Sijax.register_callback`
    :return: string - javascript code you need to put on your page,
                      to make your form use Sijax when submitted
    """

    public_name = func_name_by_form_id(form_id)

    options = _prepare_options(sijax_instance, options)
    sijax_instance.register_callback(public_name, callback, **options)

    js_params = json.dumps({PARAM_FORM_ID: form_id, PARAM_CALLBACK: public_name})
    return 'jQuery(function() { sjxUpload.registerForm(%s); });' % js_params


class UploadResponse(StreamingIframeResponse):
    """Class used for Upload handler functions,
    instead of the :class:`sijax.response.BaseResponse` class.

    This class extends :class:`sijax.response.BaseResponse` and
    every available method from it works here too.
    """

    def __init__(self, *args, **kwargs):
        StreamingIframeResponse.__init__(self, *args, **kwargs)

        # The form_id that we're dealing with is the only expected argument.
        req_args = self._request_args
        req_args_len = len(req_args)
        if req_args_len != 1:
            raise RuntimeError('We only expect 1 request argument, %d given.' %
                               req_args_len)
        self._form_id = req_args[0]

        # Let's override the request arguments now
        # The upload handler function is to expect only one argument
        # which is a dictionary of form values
        # To get the form values from all the POST data,
        # we simply need to get rid of some internal/system params

        # We're creating a new dict, because the POST data dictionary
        # may be immutable and/or we don't want to change it anyways.
        # We want to work on a copy of it.
        form_values = dict(self._sijax.get_data())
        del form_values[self._sijax.__class__.PARAM_REQUEST]
        del form_values[self._sijax.__class__.PARAM_ARGS]
        self._request_args = [form_values]

    @property
    def form_id(self):
        """The id of the form that was submitted.

        You can have several forms handled by the same upload handler
        function, which can be confusing unless you know the form id.
        """
        return self._form_id

    def reset_form(self):
        """Resets the form to the state it was in at page loading time.

        Any changes made to it after that will be cleared."""
        return self.call('sjxUpload.resetForm', [self.form_id])



# -*- coding: utf-8 -*-

"""
    sijax.response.base
    ~~~~~~~~~~~~~~~~~~~

    Provides the BaseResponse class, which is the base class
    used to create response objects for Sijax functions.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from ..helper import json
from ..exception import SijaxError
from types import GeneratorType


class BaseResponse(object):
    """The response class is the way by which Sijax functions (handlers)
    pass information back to the browser. They do this by calling
    various methods, which queue commands until they're sent to the browser.
    """
    
    COMMAND_ALERT = 'alert'
    COMMAND_HTML = 'html'
    COMMAND_SCRIPT = 'script'
    COMMAND_ATTR = 'attr'
    COMMAND_CSS = 'css'
    COMMAND_REMOVE = 'remove'
    COMMAND_CALL = 'call'
    
    def __init__(self, sijax_instance, request_args):
        """Constructs a new empty Sijax Response object.

        The arguments that the sijax function is invoked with are passed,
        so that we can override them if we want.
        We don't need to do that for regular sijax responses,
        but certain plugins may need to do that.

        :param sijax_instance: the Sijax instance object
                               that created this response
        :param request_args: the request arguments that the sijax function
                             was invoked with
        """
        self._commands = []
        self._sijax = sijax_instance
        self._request_args = request_args
    
    def _get_request_args(self):
        """Returns the arguments list to pass to callbacks.

        This class doesn't manipulate them, so it returns the original
        call arguments, but other classes can override them however they need.
        """
        return self._request_args
        
    def _add_command(self, cmd_type, params = None):
        """Adds a raw command to the buffer to send to the client."""
        if params is None:
            params = {}
        params['type'] = cmd_type
        
        self._commands.append(params)
        return self
    
    def clear_commands(self):
        """Clears the commands buffer.

        All the commands added by other methods will be removed.

        Example::

            obj_response.alert('Some alert!')
            obj_response.clear_commands()
            # The alert() above got removed from the commands queue
        """
        self._commands = []
        return self

    def alert(self, message):
        """Sends a ``window.alert`` command to the browser.

        Example that shows a message box::

            obj_response.alert('Alert message!')
        """
        params = {self.__class__.COMMAND_ALERT: message}
        return self._add_command(self.__class__.COMMAND_ALERT, params)
    
    def _html(self, selector, html, set_type):
        params = {'selector': selector, 'html': html, 'setType': set_type}
        return self._add_command(self.__class__.COMMAND_HTML, params)
   
    def html(self, selector, html):
        """Assigns the given html value to all elements
        matching the jQuery selector.

        Scripts inside the html block are also executed.

        Same as jQuery's: ``$(selector).html(value)``

        Example which sets a new html value for an element::

            obj_response.html('#element', '<strong>Hey!</strong>')

        :param selector: the jQuery selector for which we'll replace the html
        :param html: the html text
        """
        return self._html(selector, html, 'replace')

    def html_append(self, selector, html):
        """Same as :meth:`sijax.response.BaseResponse.html`,
        but appends instead of assigning a new value.
        """
        return self._html(selector, html, "append")
   
    def html_prepend(self, selector, html):
        """Same as :meth:`sijax.response.BaseResponse.html`,
        but prepends instead of assigning a new value.
        """
        return self._html(selector, html, 'prepend')
    
    def script(self, js):
        """Executes the given javascript code.

        Example::

            obj_response.html("alert('Javascript code!');")

        Note that the given javascript code is eval-ed inside a
        Sijax helper function in the browser, so it's not touching
        the global namespace, unless you explicitly do it.
        """
        params = {self.__class__.COMMAND_SCRIPT: js}
        return self._add_command(self.__class__.COMMAND_SCRIPT, params)

    def css(self, selector, property_name, value):
        """Assigns a style property value to all elements
        matching the jQuery selector.

        Same as jQuery's ``$(selector).css('property', 'value')``

        Example which changes the background color of an element::

            obj_response.css('#element', 'backgroundColor', 'red')

        Note that this can only change a single property at once.
        It doesn't support jQuery's mass change which uses a
        key/value map of properties/values to change.

        :param selector: the jQuery selector to invoke css() on
        :param property_name: the name of the style property
        :param value: the new value to assign to the property
        """
        params = {'selector': selector, 'key': property_name, 'value': value}
        return self._add_command(self.__class__.COMMAND_CSS, params)

    def _attr(self, selector, property_name, value, set_type):
        params = {
            'selector': selector, 'key': property_name,
            'value': value, 'setType': set_type
        }
        return self._add_command(self.__class__.COMMAND_ATTR, params)

    def attr(self, selector, property_name, value):
        """Assigns an attribute value to all elements
        matching the jQuery selector.

        Same as jQuery's ``$(selector).attr('property', 'value')``

        Example which changes 2 attributes of a given element::

            obj_response.attr('#element', 'width', '500px')
            obj_response.attr('#element', 'disabled', true)

        :param selector: the jQuery selector to invoke attr() on
        :param property_name: the name of the property
        :param value: the new value to assign to the property
        """
        return self._attr(selector, property_name, value, 'replace')
    
    def attr_append(self, selector, property_name, value):
        """Same as :meth:`sijax.response.BaseResponse.attr`,
        but appends instead of assigning a new value."""
        return self._attr(selector, property_name, value, 'append')

    def attr_prepend(self, selector, property_name, value):
        """Same as :meth:`sijax.response.BaseResponse.attr`,
        but prepends instead of assigning a new value."""
        return self._attr(selector, property_name, value, 'prepend')
    
    def remove(self, selector):
        """Removes all elements that match the jQuery selector from the DOM.

        Example which removes all DIV elements from the page::

            obj_response.remove('div')

        Same as jQuery's: ``$(selector).remove()``
        """
        params = {self.__class__.COMMAND_REMOVE: selector}
        return self._add_command(self.__class__.COMMAND_REMOVE, params)

    
    def redirect(self, uri):
        """Redirects the browser to the given URI.

        Example::

            obj_response.redirect('http://example.com/')

        """
        return self.script('window.location = %s;' % json.dumps(uri))
    
    def call(self, js_func_name, func_params=None):
        """Calls the given javascript function with the given arguments list.

        Example which calls the browser's ``alert()`` function::

            obj_response.call('alert', ['Message'])

        :param js_func_name: the name of the javascript function to call
        :param func_params: a list of arguments to call the function with
        """
        if func_params is None:
            func_params = []

        if (not isinstance(func_params, list) and
            not isinstance(func_params, tuple)):
            raise SijaxError('call() expects a list, a tuple or '
                             'None for the args list')

        params = {
            self.__class__.COMMAND_CALL: js_func_name,
            'params': func_params
        }
        return self._add_command(self.__class__.COMMAND_CALL, params)
    
    def _get_json(self):
        """Returns a JSON representation of the commands buffer list.

        The client side code will loop over the list and execute all the
        commands in order.
        """
        return json.dumps(self._commands)
    
    def _process_callback(self, callback, args):
        """Processes a single callback.

        For normal responses this only means calling the callback,
        because we'll flush the commands to the client once (in the end).

        For other responses (like Comet) though we can flush implicitly after
        every callback.
        """

        try:
            result = callback(self, *args)
        except TypeError:
            # we should re-raise the exception if it's coming
            # from within the function, meaning calling works.. @todo
            event_invalid_call = self._sijax.__class__.EVENT_INVALID_CALL
            self._sijax.get_event(event_invalid_call)(self, callback)
        else:
            # We usually don't expect a return value,
            # but if we get a generator, it may mean that our regular function
            # was used like a streaming function (yield).
            # This is a a mistake which could happen due to incorrect
            # function registration.
            if isinstance(result, GeneratorType):
                raise SijaxError('Flushing/Yielding/Streaming is not '
                                 'supported for regular functions!')

    def _process_call_chain(self, call_chain):
        """Executes all the callbacks in the chain for this response object.

        Each callback in the chain adds commands to the buffer.
        When all the callbacks have been executed, the buffer would contain
        a list of commands that we need to pass to the browser (in order).

        The result of this is a JSON string to pass to the browser.

        :param call_chain: a list of two-tuples (callback, args list) to call
        """
        for callback, args in call_chain:
            self._process_callback(callback, args)
        return self._get_json()




# -*- coding: utf-8 -*-

from . import plugin
from .core import Sijax

__title__ = 'sijax'
__version__ = '0.2.0'
__author__ = 'Slavi Pantaleev'
__license__ = 'BSD'
__copyright__ = 'Copyright 2011 Slavi Pantaleev'




# -*- coding: utf-8 -*

"""
    sijax.core
    ~~~~~~~~~~

    Implements the main Sijax class, an instance of which is used to
    register callbacks, invoke callbacks, setup events handlers, etc.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from .helper import json

from .response.base import BaseResponse
from .exception import SijaxError

class Sijax(object):
    """The main Sijax object is what manages function registration and calling.

    Sijax initialization usually looks like this::

        instance = Sijax()
        instance.set_data(POST_DICTIONARY_HERE)
        instance.set_request_uri(URI_TO_SEND_REQUESTS_TO)
        instance.register_callback('function_name', some_function)

    Sijax needs the POST parameters for the current request, to determine
    whether the request is meant to be handled by Sijax or by your regular
    page loading logic.

    Sijax needs a request URI to tell the client where to
    send the ajax requests.
    This is usually the current request URI, but another URI can also be used.

    Functions that are registered (using :meth:`sijax.Sijax.register_callback`
    or :meth:`sijax.Sijax.register_object`) are the only functions exposed to the
    browser for calling.
    """
    
    PARAM_REQUEST = 'sijax_rq'
    PARAM_ARGS = 'sijax_args'
    
    #: Event called immediately before calling the response function.
    #: The event handler function receives the Response object argument.
    EVENT_BEFORE_PROCESSING = 'before_processing'

    #: Event called immediately after calling the response function.
    #: The event handler function receives the Response object argument.
    EVENT_AFTER_PROCESSING = 'after_processing'

    #: Event called when the function to be called is unknown
    #: (not registered with Sijax).
    #: The event handler function receives the Response object argument
    #: followed by the public name of the function that was
    #: supposed to be called.
    EVENT_INVALID_REQUEST = 'invalid_request'
    
    #: Event called when the function was called in a wrong way
    #: (bad arguments count).
    #: The event handler function receives the Response object argument
    #: followed by the callback function that was not called correctly.
    EVENT_INVALID_CALL = 'invalid_call'

    #: An option when registering callbacks that stores the callback function
    PARAM_CALLBACK = 'callback'

    #: An option when registering callbacks that allows the response class to
    #: be changed from :class:`sijax.response.BaseResponse` to something else.
    PARAM_RESPONSE_CLASS = 'response_class'

    #: An option when registering callbacks that makes certain extra arguments
    #: be passed automatically to the response function after the
    #: obj_response argument and before the other arguments
    PARAM_ARGS_EXTRA = 'args_extra'
    
    def __init__(self):
        cls = self.__class__

        #: Would contain the request data (usually POST)
        self._data = {}
        
        #: Would contain the callbacks (name => dictionary of params)
        self._callbacks = {}

        #: Stores the various event callbacks
        self._events = {}

        #: The URI where the client should send ajax requests to
        self._request_uri = None

        #: The URI to json2.js or similar to provide JSON support
        #: for browsers that don't have it.. Sijax would load this on demand
        self._json_uri = None
        
        #: Stores a cached version of the arguments
        #: to be passed to the requested function
        self._request_args = None

        def invalid_request(obj_response, func_name):
            """Handler to be called when an unknown function is called."""

            msg = 'The action you performed is unavailable! (Sijax error)'
            obj_response.alert(msg)

        def invalid_call(obj_response, callback):
            """Handler to be called when a function is called the wrong way."""

            msg = 'You tried to perform an action in a wrong way! (Sijax error)'
            obj_response.alert(msg)

        # Register the before/after processing events
        # to not do anything by default
        self.register_event(cls.EVENT_BEFORE_PROCESSING, lambda r: r)
        self.register_event(cls.EVENT_AFTER_PROCESSING, lambda r: r)

        # Register the "error" events to show some alerts by default
        self.register_event(cls.EVENT_INVALID_REQUEST, invalid_request)
        self.register_event(cls.EVENT_INVALID_CALL, invalid_call)
    
    def set_data(self, data):
        """Sets the incoming data dictionary (usually POST).

        Sijax needs this data to determine if the current request
        is a Sijax request, and which callback should be called
        with what arguments.
        """
        self._data = data
        self._request_args = None # cache invalidation
        return self

    def get_data(self):
        """Returns the incoming data array that the current instance uses."""
        return self._data
    
    def register_callback(self, public_name, callback, response_class=None,
                          args_extra=None):
        """Registers the specified callback function with the given name.

        Example::

            def handler(obj_response):
                pass

            # Exposing the same function by 2 different names
            instance.register_callback('test', handler)
            instance.register_callback('test2', handler)

        The optional ``response_class`` parameter could be used to substitute the
        :class:`sijax.response.BaseResponse` class used by default.
        An instance of ``response_class`` is passed automatically
        as a first parameter to your response function (``obj_response``).

        Example::

            def handler(obj_response):
                pass

            class MyResponse(sijax.response.BaseResponse):
                pass

            # `obj_response` passed to `handler` will be an instance of MyResponse
            instance.register_callback('test', handler, response_class=MyResponse)

        The optional ``args_extra`` parameter allows you to pass a list of
        extra arguments to your response function, immediately after
        the ``obj_response`` argument and before the other call arguments.

        Example::

            def handler(obj_response, arg1, arg2, arg3):
                # arg1 = 'arg1'
                # arg2 = 'arg2'
                # arg3 - expected to come from the browser
                pass

            instance.register_callback('test', handler, args_extra=['arg1', 'arg2'])

        :param public_name: the name with which this function will be
                            exposed in the browser
        :param callback: the actual function to call
        :param response_class: the response class, an instance of which to use
                               instead of :class:`sijax.response.BaseResponse`
        :param args_extra: an optional list of additional arguments to pass
                           after the response object argument and before the
                           other call arguments
        """

        if response_class is None:
            response_class = BaseResponse
        if not hasattr(response_class, '__call__'):
            raise SijaxError('response_class needs to be a callable!')

        if args_extra is None:
            args_extra = []
        else:
            if isinstance(args_extra, tuple):
                args_extra = list(args_extra)
            elif not isinstance(args_extra, list):
                raise SijaxError('args_extra could only be a tuple or a list!')

        params = {}
        params[self.__class__.PARAM_CALLBACK] = callback
        params[self.__class__.PARAM_RESPONSE_CLASS] = response_class
        params[self.__class__.PARAM_ARGS_EXTRA] = args_extra

        self._callbacks[public_name] = params
        return self
    
    def register_object(self, obj, **options):
        """Registers all "public" callable attributes of the given object.
        
        The object could be anything (module, class, class instance, etc.)

        :param obj: the object whose callable attributes to register
        :param options: the options to be sent to
                        :meth:`sijax.Sijax.register_callback`
        """
        for attr_name in dir(obj):
            if attr_name.startswith('_'):
                continue
            
            attribute = getattr(obj, attr_name)
            if hasattr(attribute, '__call__'):
                self.register_callback(attr_name, attribute, **options)
    
    @property
    def is_sijax_request(self):
        """Tells whether this page request looks like
        a valid request meant to be handled by Sijax.

        Even if this is a Sijax request, this doesn't mean it's a valid one.
        Refer to :meth:`sijax.Sijax.process_request` to see what happens when
        a request for an unknown function was made.
        """
        for k in (self.__class__.PARAM_REQUEST, self.__class__.PARAM_ARGS):
            if k not in self._data:
                return False
        return True
    
    @property
    def requested_function(self):
        """The name of the requested function,
        or None if the current request is not a Sijax requset."""
        request = self.__class__.PARAM_REQUEST
        return str(self._data[request]) if request in self._data else None
    
    @property
    def request_args(self):
        """The arguments list, that the function was called with
        in the browser.

        It should be noted that this is not necessarily the arguments list
        that the callback function will actually receive.
        Custom Response objects are allowed to override the arguments list.
        """
        key_args = self.__class__.PARAM_ARGS
        if self._request_args is None:
            # no precached version.. cache it now
            self._request_args = []
            if key_args in self._data:
                try:
                    args = json.loads(self._data[key_args])
                    if isinstance(args, list):
                        self._request_args = args
                except (ValueError):
                    pass
        return self._request_args
    
    def process_request(self):
        """Executes the Sijax request and returns the response.

        Make sure that the current request is a Sijax request,
        using :attr:`sijax.Sijax.is_sijax_request` before calling this.
        A :class:`sijax.exception.SijaxError` will be raised,
        if this method is called for non-Sijax requests.

        If the function that was called from the browser is not
        registered (the function is unknown to Sijax),
        the :attr:`sijax.Sijax.EVENT_INVALID_REQUEST` event handler will
        be called instead.

        Refer to :meth:`sijax.Sijax.execute_callback` to see how the main
        handler is called and what the response (return value) is.
        """

        if not self.is_sijax_request:
            raise SijaxError('You should not call this for non-Sijax requests!')
        
        function_name = self.requested_function
        if function_name in self._callbacks:
            options = self._callbacks[function_name]
            args = self.request_args
        else:
            # Function not registered.. Let's call the invalid request handler
            # passing to it the function name that should've been called
            args = [function_name]
            callback = self._events[self.__class__.EVENT_INVALID_REQUEST]
            options = {self.__class__.PARAM_CALLBACK: callback}
        
        return self.execute_callback(args, **options)
    
    def execute_callback(self, args, callback, **params):
        """Executes the given callback function and returns a response.

        Before executing the given callback, the
        :attr:`sijax.Sijax.EVENT_BEFORE_PROCESSING` event callback is fired.
        After executing the given callback, the 
        :attr:`sijax.Sijax.EVENT_AFTER_PROCESSING` event callback is fired.

        The returned result could either be a string (for regular functions)
        or a generator (for streaming functions like Comet or Upload).
        
        :param args: the arguments list to pass to the callback
        :param callback: the callback function to execute
        :param options: more options - look at
                                       :meth:`sijax.Sijax.register_callback`
                                       to see what else is available
        :return: string for regular callbacks or generator for streaming callbacks
        """

        cls = self.__class__

        # Another response class could be used to extend behavior
        response_class = BaseResponse
        if cls.PARAM_RESPONSE_CLASS in params:
            response_class = params[cls.PARAM_RESPONSE_CLASS]

        args_extra = []
        if cls.PARAM_ARGS_EXTRA in params:
            args_extra = params[cls.PARAM_ARGS_EXTRA]

        if not hasattr(callback, '__call__'):
            raise SijaxError('Provided callback is not callable!')

        # Pass the original call args to the response and get them back later
        # This allows the response class (potentially custom)
        # to override the arguments list.
        # Note that we're not passing args_extra to it, as we don't
        # want responses to know anything about that.
        obj_response = response_class(self, args)
        call_args = args_extra + obj_response._get_request_args()

        call_chain = [
            (self._events[cls.EVENT_BEFORE_PROCESSING], []),
            (callback, call_args),
            (self._events[cls.EVENT_AFTER_PROCESSING], [])
        ]
        return obj_response._process_call_chain(call_chain)

    def register_event(self, event_name, callback):
        """Register a callback function to be called when the event occurs.

        Only one callback can be executed per event.

        The provided ``EVENT_*`` constants should be used
        for handling system events.
        Additionally, you can use any string to define your
        own events and callbacks.

        If more than one handler per event is needed, you can chain them
        manually, although it's not recommended.
        """
        self._events[event_name] = callback
        return self

    def has_event(self, event_name):
        """Tells whether we've got a registered event by that name."""
        return event_name in self._events

    def get_event(self, event_name):
        """Returns the event handler callback for the specified event."""
        return self._events[event_name] if event_name in self._events else None

    def set_request_uri(self, uri):
        """Specifies the URI where ajax requests will be sent to
        by the client.

        This is usually the URI of the current request.
        """
        self._request_uri = uri
        return self
    
    def set_json_uri(self, uri):
        """Sets the URI to an external JSON library,
        for browsers that do not support native JSON (such as IE <= 7).

        If this is not specified, Sijax will not work at all in such browsers.

        The script will only be loaded if a browser
        without JSON support is detected.
        """
        self._json_uri = uri
        return self
    
    def get_js(self):
        """Returns the javascript code needed to prepare the
        Sijax environment in the browser.

        Note that the javascript code is unique and cannot be shared between
        different pages.
        """

        if self._request_uri is None:
            raise SijaxError('Trying to get the sijax js, '
                             'but no request_uri has been set!')
        
        js = 'Sijax.setRequestUri(%s);' % json.dumps(self._request_uri)
        if self._json_uri is not None:
            js += 'Sijax.setJsonUri(%s);' % json.dumps(self._json_uri) 
        return js




from __future__ import with_statement

import sys
import os
import unittest
import tempfile
import shutil
from contextlib import contextmanager


sijax_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(sijax_path)

from sijax import Sijax
from sijax.response import BaseResponse, StreamingIframeResponse
from sijax.exception import SijaxError

from sijax.plugin.comet import register_comet_callback, register_comet_object, \
     CometResponse

from sijax.plugin.upload import register_upload_callback, UploadResponse

from sijax.helper import init_static_path

@contextmanager
def temporary_dir(*args, **kwargs):
    path = tempfile.mkdtemp()
    try:
        yield path
    finally:
        shutil.rmtree(path)


class SijaxMainTestCase(unittest.TestCase):

    def test_changing_incoming_data_works(self):
        inst = Sijax()
        self.assertTrue(inst.get_data() == {})

        data = {"key": "value"}
        inst.set_data(data)
        self.assertTrue(inst.get_data() == data)

        data = {"key": "value", "another": "one"}
        self.assertTrue(inst.get_data() == {"key": "value"})
        inst.set_data(data)
        self.assertTrue(inst.get_data() == data)

    def test_detecting_sijax_requests_works(self):
        inst = Sijax()
        cls = inst.__class__

        # no POST data specified.. this surely is a GET request
        self.assertFalse(inst.is_sijax_request)
        self.assertTrue(inst.requested_function is None)

        # missing params that specify the function and args
        inst.set_data({"key": "value"})
        self.assertFalse(inst.is_sijax_request)
        self.assertTrue(inst.requested_function is None)
        
        # missing args
        inst.set_data({cls.PARAM_REQUEST: "function"})
        self.assertFalse(inst.is_sijax_request)
        self.assertEqual(inst.requested_function, "function")

        # missing function name
        inst.set_data({cls.PARAM_ARGS: "[]"})
        self.assertFalse(inst.is_sijax_request)
        self.assertTrue(inst.requested_function is None)
        self.assertEqual([], inst.request_args)

        inst.set_data({cls.PARAM_ARGS: '["invalid_json": here'})
        self.assertFalse(inst.is_sijax_request)
        self.assertTrue(inst.requested_function is None)
        self.assertEqual([], inst.request_args)

        # this should be considered valid.. both params available
        # the arguments are json encoded
        inst.set_data({cls.PARAM_REQUEST: "function", cls.PARAM_ARGS: '["arg1", 5]'})
        self.assertTrue(inst.is_sijax_request)
        self.assertEqual("function", inst.requested_function)
        self.assertEqual(["arg1", 5], inst.request_args)

        inst.set_data({cls.PARAM_REQUEST: "func2", cls.PARAM_ARGS: '[28, 5]'})
        self.assertTrue(inst.is_sijax_request)
        self.assertEqual("func2", inst.requested_function)
        self.assertEqual([28, 5], inst.request_args)

    def test_events_have_default_handlers(self):
        inst = Sijax()
        cls = inst.__class__
        events = [getattr(cls, attr) for attr in dir(cls) if attr.startswith('EVENT_')]
        for event in events:
            self.assertTrue(inst.has_event(event), "No handler for %s" % event)

    def test_registering_custom_events_works(self):
        inst = Sijax()

        event_name = "my_event"
        event_callback = lambda obj_response: obj_response.alert("Event")

        self.assertFalse(inst.has_event(event_name), "Custom event registered")

        inst.register_event(event_name, event_callback)
        
        self.assertTrue(inst.has_event(event_name), "Failed to register event")
                
    def test_executing_regular_callbacks_works(self):
        from types import StringType

        calls_history = []
        call_args_history = []

        def event_before(obj_response):
            self.assertTrue(isinstance(obj_response, BaseResponse))
            calls_history.append("before")

            obj_response.script("javascript here..")

        def event_after(obj_response):
            self.assertTrue(isinstance(obj_response, BaseResponse))
            calls_history.append("after")

            obj_response.css("#element", "backgroundColor", "red")

        def callback_main(obj_response, arg1, arg2):
            self.assertTrue(isinstance(obj_response, BaseResponse))

            calls_history.append("main")
            call_args_history.append(arg1)
            call_args_history.append(arg2)

            obj_response.alert("alert from main")

        inst = Sijax()
        cls = inst.__class__
        inst.register_event(cls.EVENT_BEFORE_PROCESSING, event_before)
        inst.register_event(cls.EVENT_AFTER_PROCESSING, event_after)

        call_args = ["arg1", 15]
        response = inst.execute_callback(call_args, callback=callback_main)

        self.assertEqual(["before", "main", "after"], calls_history)
        self.assertEqual(call_args, call_args_history)

        # the response should be a string for regular functions
        # streaming functions return generators instead..
        self.assertTrue(isinstance(response, StringType))

        from sijax.helper import json
        try:
            commands = json.loads(response)
        except:
            self.fail("Invalid JSON response!")
        else:
            self.assertTrue(isinstance(commands, list))

            commands_history = []
            for cmd_params in commands:
                self.assertTrue(isinstance(cmd_params, dict))
                self.assertTrue("type" in cmd_params, "Unknown command type!")
                commands_history.append(cmd_params["type"])

        self.assertEqual(["script", "alert", "css"], commands_history)

    def test_bad_callback_objects_raise_exception(self):
        inst = Sijax()
        try:
            inst.execute_callback(callback='non-callable object', args=[])
            self.fail('SijaxError not raised when bad callback was given!')
        except SijaxError:
            pass

    def test_get_js_fails_with_missing_request_uri(self):
        inst = Sijax()
        try:
            inst.get_js()
            self.fail()
        except SijaxError:
            pass

    def test_changing_uris_results_in_a_differnt_js_output(self):
        inst = Sijax()

        req_uri = "http://localhost:8080/submit_here"
        json_uri = "http://localhost:8080/json2.js"

        inst.set_request_uri(req_uri)
        js = inst.get_js()
        self.assertTrue('Sijax.setRequestUri("%s");' % req_uri in js)
        self.assertFalse('Sijax.setJsonUri' in js)

        inst.set_json_uri(json_uri)
        js = inst.get_js()
        self.assertTrue('Sijax.setRequestUri("%s");' % req_uri in js)
        self.assertTrue('Sijax.setJsonUri("%s");' % json_uri in js)

    def test_process_request_throws_exception_when_called_for_non_sijax_requests(self):
        inst = Sijax()
        try:
            inst.process_request()
            self.fail("Process request got executed for a non-sijax request!")
        except SijaxError:
            pass

    def test_process_request_calls_invalid_request_event_for_invalid_requests(self):
        from types import StringType
        from sijax.helper import json

        # An invalid request is a request for a function that's not registered,
        # meaning the request is invalid as far as sijax is concerned
        inst = Sijax()
        cls = inst.__class__
        inst.set_data({cls.PARAM_REQUEST: "my_func", cls.PARAM_ARGS: '["arg1", 12]'})
        self.assertTrue(inst.is_sijax_request)
        self.assertEqual("my_func", inst.requested_function)
        self.assertEqual(["arg1", 12], inst.request_args)

        response = inst.process_request()
        self.assertTrue(isinstance(response, StringType))

        try:
            commands = json.loads(response)
        except:
            self.fail("Invalid JSON generated!")
        else:
            self.assertTrue(isinstance(commands, list))

            # we expect the default "Invalid request" alert
            self.assertEqual(1, len(commands))
            command_data = commands.pop(0)
            self.assertTrue("type" in command_data)
            self.assertEqual("alert", command_data["type"])

    def test_process_request_calls_invalid_call_event_for_invalid_calls(self):
        from types import StringType, FunctionType
        from sijax.helper import json

        # An invalid call is a call to a function that appears valid.
        # The function is registered (known), but calling fails, because
        # the function expects another number of arguments

        call_history = []

        def my_callback(obj_response, arg1, arg2):
            call_history.append("call ok")
        
        def my_callback_with_defaults(obj_response, arg1=138, arg2=15):
            call_history.append("defaults ok")

        def invalid_call(obj_response, failed_callback):
            self.assertTrue(isinstance(failed_callback, FunctionType))
            call_history.append("invalid %s" % failed_callback.__name__)

        inst = Sijax()
        cls = inst.__class__
        inst.register_event(cls.EVENT_INVALID_CALL, invalid_call)
        inst.register_callback("my_func", my_callback)

        inst.set_data({cls.PARAM_REQUEST: "my_func", cls.PARAM_ARGS: '["arg1", 12]'})
        self.assertTrue(inst.is_sijax_request)
        self.assertEqual("my_func", inst.requested_function)
        self.assertEqual(["arg1", 12], inst.request_args)
        response = inst.process_request()
        self.assertTrue(isinstance(response, StringType))

        # we should have succeeded now..
        # let's try to make the call invalid and observe the failure
        inst.set_data({cls.PARAM_REQUEST: "my_func", cls.PARAM_ARGS: '[]'})
        self.assertEqual("my_func", inst.requested_function)
        self.assertEqual([], inst.request_args)
        response = inst.process_request()
        self.assertTrue(isinstance(response, StringType))

        # let's see if calling works with default arguments
        inst.register_callback("my_func", my_callback_with_defaults)
        response = inst.process_request()
        self.assertTrue(isinstance(response, StringType))

        self.assertEqual(["call ok", "invalid my_callback", "defaults ok"], call_history)

    def test_new_callbacks_override_old_during_registering(self):
        call_history = []

        def callback_one(obj_response):
            call_history.append("one")

        def callback_two(obj_response):
            call_history.append("two")

        inst = Sijax()
        cls = inst.__class__
        inst.set_data({cls.PARAM_REQUEST: "my_func", cls.PARAM_ARGS: '[]'})
        self.assertTrue(inst.is_sijax_request)

        inst.register_callback("my_func", callback_one)
        inst.process_request()

        inst.register_callback("my_func", callback_two)
        inst.process_request()

        self.assertEqual(["one", "two"], call_history)

    def test_mass_registering_from_a_class_works(self):
        call_history = []

        class SijaxHandler(object):
            @staticmethod
            def callback_one(obj_response):
                call_history.append("one")

            @staticmethod
            def callback_two(obj_response):
                call_history.append("two")

            @classmethod
            def callback_three(cls, obj_response):
                call_history.append("three")

        inst = Sijax()
        inst.register_object(SijaxHandler)
        cls = inst.__class__

        inst.set_data({cls.PARAM_REQUEST: "callback_one", cls.PARAM_ARGS: '[]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        inst.set_data({cls.PARAM_REQUEST: "callback_two", cls.PARAM_ARGS: '[]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        inst.set_data({cls.PARAM_REQUEST: "callback_three", cls.PARAM_ARGS: '[]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        self.assertEqual(["one", "two", "three"], call_history)

    def test_mass_registering_from_an_object_works(self):
        call_history = []

        class SijaxHandler(object):
            def callback_one(self, obj_response):
                call_history.append("one")

            def callback_two(self, obj_response):
                call_history.append("two")

        inst = Sijax()
        inst.register_object(SijaxHandler())
        cls = inst.__class__

        inst.set_data({cls.PARAM_REQUEST: "callback_one", cls.PARAM_ARGS: '[]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        inst.set_data({cls.PARAM_REQUEST: "callback_two", cls.PARAM_ARGS: '[]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        self.assertEqual(["one", "two"], call_history)

    def test_the_response_class_could_be_changed_during_registration(self):
        # Sometimes it may be convenient for the developer to be able to
        # change the class behind obj_response on a per-callback basis
        call_history = []

        class CustomResponse(BaseResponse): pass

        def my_callback(obj_response):
            call_history.append("regular")
            self.assertFalse(isinstance(obj_response, CustomResponse))
            self.assertTrue(isinstance(obj_response, BaseResponse))

        def my_custom_callback(obj_response):
            call_history.append("custom")
            self.assertTrue(isinstance(obj_response, CustomResponse))


        inst = Sijax()
        cls = inst.__class__

        inst.register_callback("my_func", my_callback)
        params_custom = {cls.PARAM_RESPONSE_CLASS: CustomResponse}
        inst.register_callback("my_func_custom", my_custom_callback, **params_custom)

        inst.set_data({cls.PARAM_REQUEST: "my_func", cls.PARAM_ARGS: '[]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        inst.set_data({cls.PARAM_REQUEST: "my_func_custom", cls.PARAM_ARGS: '[]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        self.assertEqual(["regular", "custom"], call_history)

    def test_args_extra_works_properly(self):
        call_history = []

        def callback(obj_response, arg1_custom, arg2_custom, arg_regular):
            self.assertTrue(isinstance(obj_response, BaseResponse))
            call_history.append(arg1_custom)
            call_history.append(arg2_custom)
            call_history.append(arg_regular)
            
        def callback_basic(obj_response, arg_regular):
            self.assertTrue(isinstance(obj_response, BaseResponse))
            call_history.append(arg_regular)

        inst = Sijax()
        cls = inst.__class__

        inst.register_callback("callback", callback, args_extra=("one", "two"))
        inst.set_data({cls.PARAM_REQUEST: "callback", cls.PARAM_ARGS: '["regular"]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        inst.register_callback("callback", callback_basic)
        inst.set_data({cls.PARAM_REQUEST: "callback", cls.PARAM_ARGS: '["reg2"]'})
        self.assertTrue(inst.is_sijax_request)
        inst.process_request()

        call_history_expected = ["one", "two", "regular", "reg2"]
        self.assertEqual(call_history_expected, call_history)
    
    def test_args_extra_expects_a_list_or_a_tuple(self):
        inst = Sijax()

        def try_args(args_extra, should_succeed):
            success = False
            try:
                inst.register_callback("name", lambda r: r, args_extra=args_extra)
                success = True
            except SijaxError:
                pass

            self.assertEqual(should_succeed, success, "Failure for %s" % repr(args_extra))
                
        # A list, a tuple and None should all work
        try_args([], True)
        try_args((), True)
        try_args(None, True)
        try_args(['one item'], True)
        try_args(('one item',), True)
        try_args(['one', 'two'], True)
        try_args(('one', 'two'), True)
        try_args([{"dictionary": "in", "the": "list"}], True)
        try_args(({"dictionary": "in", "the": "tuple"},), True)

        try_args('string', False)
        try_args(True, False)
        try_args(False, False)
        try_args({"dictionary": "here"}, False)

    def test_regula_functions_that_yield_are_not_allowed(self):
        # Yielding is only supported by streaming functions.
        # It makes no sense for regular functions to use it.
        # If a regular function tries to yield, we expect
        # a SijaxError to be raised

        inst = Sijax()
        def callback(obj_response):
            yield obj_response

        try:
            inst.execute_callback([], callback)
            self.fail("Yielding regular function didn't raise expected exception!")
        except SijaxError:
            pass

    def test_response_method_call_rejects_bad_args(self):
        # obj_response.call receives a function name to call
        # and an optional list of arguments to pass to it.
        # If the arguments parameter is specified, it must be
        # a list or a tuple.

        inst = Sijax()

        def try_call(args_value, should_succeed):
            def callback(obj_response):
                obj_response.call('function', args_value)

            success = False
            try:
                inst.execute_callback([], callback)
                success = True
            except SijaxError, ex:
                pass

            self.assertEqual(should_succeed, success, "Failure for %s" %
                             repr(args_value))

        try_call(None, True)
        try_call([], True)
        try_call(['arg1', 'arg2', 3, 4, 'arg5'], True)
        try_call((), True)

        try_call('arg1', False)
        try_call(1, False)
        try_call({'dictionary': 'here'}, False)

    def test_init_static_path_helper_works(self):
        import os
        import sijax

        with temporary_dir() as static_path:
            init_static_path(static_path)

            version_file = os.path.join(static_path, 'sijax_version')
            if not os.path.exists(version_file):
                self.fail('Version file %s does not exist' % version_file)
            self.assertEqual(open(version_file).read(), sijax.__version__)

            core_js_file = os.path.join(static_path, 'sijax.js')

            # let's ensure that files are only written (or deleted)
            # if the version changes
            fp = open(core_js_file, 'w')
            fp.write('new stuff')
            fp.close()
            init_static_path(static_path)
            # version file is the same, so it shouldn't touch it
            self.assertEqual(open(core_js_file).read(), 'new stuff')


            fp = open(version_file, 'w')
            fp.write('another_version_string')
            fp.close()

            # let's also create another file (as it's from the old version)
            # and make sure that init will delete it
            new_file = os.path.join(static_path, 'extra-file.js')
            self.assertFalse(os.path.exists(new_file))
            fp = open(new_file, 'w')
            fp.write('blah')
            fp.close()
            self.assertTrue(os.path.exists(new_file))

            # the version string is different, so we expect a complete resync
            # - extra files should be deleted, other files should be updated
            init_static_path(static_path)
            self.assertNotEqual(open(core_js_file).read(), 'new stuff')
            self.assertEqual(open(version_file).read(), sijax.__version__)
            self.assertFalse(os.path.exists(new_file))

    def test_static_path_helper_refuses_to_write_to_non_empty_paths(self):
        # The static path helper should only write to paths that are empty
        # or that include a version file.
        # Only such paths are considered safe to write to.
        import os

        def try_init(path, should_succeed):
            success = False
            try:
                init_static_path(path)
                success = True
            except SijaxError:
                pass

            self.assertEqual(should_succeed, success)

        # New empty temporary dir should work
        with temporary_dir() as static_path:
            try_init(static_path, True)

        # A directory with some files (but not version file) should fail
        with temporary_dir() as static_path:
            fp = open(os.path.join(static_path, 'some.file'), 'w')
            fp.write('blah')
            fp.close()
            try_init(static_path, False)

        # A directory with some files, but also a version file
        with temporary_dir() as static_path:
            fp = open(os.path.join(static_path, 'some.file'), 'w')
            fp.write('blah')
            fp.close()
            fp = open(os.path.join(static_path, 'sijax_version'), 'w')
            fp.write('version_string')
            fp.close()
            try_init(static_path, True)


class SijaxStreamingTestCase(unittest.TestCase):
    """This tests the StreamingIframeResponse functionality, which is
    used behind the Comet and Upload plugins.
    """

    def test_streaming_functions_return_generators(self):
        # Every function registered as streaming should return a generator
        # when it's executed, even if the actual function decides to act
        # as a normal function, and not generate anything
        # This is very important, because it allows mixing streaming with
        # regular functions when the StreamingIframeResponse is used

        from types import GeneratorType

        call_history = []

        def callback_before(obj_response):
            self.assertTrue(isinstance(obj_response, StreamingIframeResponse))
            call_history.append("before")
            obj_response.html("#div", "before html")

        def callback_after(obj_response):
            self.assertTrue(isinstance(obj_response, StreamingIframeResponse))
            call_history.append("after")
            # intentionally not pushing new commands
            # which means it should not be yielding at implicitly

        def callback_yielding(obj_response):
            self.assertTrue(isinstance(obj_response, StreamingIframeResponse))
            call_history.append("yielding")
            obj_response.alert("test")
            yield obj_response
            obj_response.css("#div", "color", "red")

        def callback_normal(obj_response):
            self.assertTrue(isinstance(obj_response, StreamingIframeResponse))
            call_history.append("normal")
            obj_response.script(".. javascript here..")

        def assert_generator_entries_length(generator, length):
            items = []
            try:
                while True:
                    items.append(generator.next())
            except StopIteration:
                pass

            self.assertEqual(length, len(items))

        inst = Sijax()
        cls = inst.__class__
        inst.register_event(cls.EVENT_BEFORE_PROCESSING, callback_before)
        inst.register_event(cls.EVENT_AFTER_PROCESSING, callback_after)
        options = {cls.PARAM_RESPONSE_CLASS: StreamingIframeResponse}
        
        response = inst.execute_callback([], callback=callback_yielding, **options)
        self.assertTrue(isinstance(response, GeneratorType))
        # We should have the following yields:
        # 1. Implicit yield after callback_before
        # 2. Explicit yield from callback_yielding
        # 3. Implicit yield when calback_yielding ends
        # Note that callback_after should not yield implicitly,
        # because it has not pushed new commands, so it makes no sense!
        assert_generator_entries_length(response, 3)

        def callback_before_new(obj_response):
            self.assertTrue(isinstance(obj_response, StreamingIframeResponse))
            call_history.append("before_new")

            # this doesn't really push data, as we haven't added commands yet
            # flushing would be wasteful
            for i in range(10):
                yield obj_response
            
            for i in range(2):
                obj_response.alert("hey")
                yield obj_response

        inst.register_event(cls.EVENT_BEFORE_PROCESSING, callback_before_new)
        inst.register_event(cls.EVENT_AFTER_PROCESSING, lambda r: r.alert("this yields"))
        response = inst.execute_callback([], callback=callback_normal, **options)
        self.assertTrue(isinstance(response, GeneratorType))
        # We're expecting the following yields:
        # 1. Explicit yield from callback_before_new
        # 2. Explicit yield from callback_before_new
        # 3. Implicit yield after callback_normal
        # 4. Implicit yield after EVENT_AFTER_PROCESSING callback
        assert_generator_entries_length(response, 4)

        call_history_expected = [
            "before", "yielding", "after",
            "before_new", "normal"
        ]
        self.assertEqual(call_history_expected, call_history)


class SijaxCometTestCase(unittest.TestCase):
    """Exercises certain Comet specific things. Most of the functionality
    that Comet is based on is tested elsewhere (StreamingIframeResponse).
    """

    def test_registering_helper_works(self):
        from types import GeneratorType

        class CustomResponse(CometResponse): pass

        def process_request(inst):
            self.assertTrue(inst.is_sijax_request)
            response = inst.process_request()
            self.assertTrue(isinstance(response, GeneratorType))
            # something needs to "move the generator forward" if we want
            # to proceed in the call chain
            for string in response:
                pass

        call_history = []

        def invalid_call(obj_response, callback):
            self.fail("Invalid call handler triggered!")

        inst = Sijax()
        cls = inst.__class__
        inst.register_event(cls.EVENT_INVALID_CALL, invalid_call)
        inst.set_data({cls.PARAM_REQUEST: "my_func", cls.PARAM_ARGS: '[]'})

        def my_callback(obj_response):
            self.assertTrue(isinstance(obj_response, CometResponse))
            self.assertTrue(isinstance(obj_response, StreamingIframeResponse))
            call_history.append("my_callback")
            call_history.append(type(obj_response).__name__)

        register_comet_callback(inst, "my_func", my_callback)
        process_request(inst)

        # let's see if we can override the default args and the response
        # class during callback registration
        params_custom = {cls.PARAM_RESPONSE_CLASS: CustomResponse}
        register_comet_callback(inst, "my_func", my_callback, **params_custom)
        process_request(inst)

        # Try mass registering now
        class SijaxHandler(object):
            @staticmethod
            def callback_one(obj_response, arg1):
                self.assertTrue(isinstance(obj_response, CustomResponse))
                call_history.append("callback_one")
                call_history.append(type(obj_response).__name__)
                call_history.append(arg1)

        register_comet_object(inst, SijaxHandler, **params_custom)
        inst.set_data({cls.PARAM_REQUEST: "callback_one", cls.PARAM_ARGS: '[45]'})
        process_request(inst)

        call_history_expected = [
            "my_callback", "CometResponse",
            "my_callback", "CustomResponse",
            "callback_one", "CustomResponse", 45
        ]
        self.assertEqual(call_history_expected, call_history)


class SijaxUploadTestCase(unittest.TestCase):
    """Exercises certain Comet specific things. Most of the functionality
    that Comet is based on is tested elsewhere (StreamingIframeResponse).
    """

    def test_registering_helper_works(self):
        from types import GeneratorType

        from sijax.plugin.upload import func_name_by_form_id

        class CustomResponse(UploadResponse): pass

        def process_request(inst):
            self.assertTrue(inst.is_sijax_request)
            response = inst.process_request()
            self.assertTrue(isinstance(response, GeneratorType))
            # something needs to "move the generator forward" if we want
            # to proceed in the call chain
            for string in response:
                pass

        call_history = []

        def invalid_call(obj_response, callback):
            self.fail("Invalid call handler triggered!")

        inst = Sijax()
        cls = inst.__class__
        inst.register_event(cls.EVENT_INVALID_CALL, invalid_call)

        def my_callback(obj_response, files, form_values):
            self.assertTrue(isinstance(obj_response, UploadResponse))
            self.assertTrue(isinstance(obj_response, StreamingIframeResponse))
            call_history.append("my_callback")
            call_history.append(obj_response.form_id)
            call_history.append(type(obj_response).__name__)
            call_history.append(files["file"]) # reaching into the file object

            # ensure that from all the post data, only the valid (non-system)
            # stuff made it through
            self.assertEqual({"form_key": "value", "form_key2": 15}, form_values)

        def my_callback2(obj_response, custom_arg1, custom_arg2, form_values):
            self.assertTrue(isinstance(obj_response, UploadResponse))
            self.assertTrue(isinstance(obj_response, StreamingIframeResponse))
            call_history.append("custom_callback")
            call_history.append(obj_response.form_id)
            call_history.append(type(obj_response).__name__)
            call_history.append(custom_arg1)
            call_history.append(custom_arg2)

            # ensure that from all the post data, only the valid (non-system)
            # stuff made it through
            self.assertEqual({"form_key": "value", "form_key2": 15}, form_values)

        form_id = "my_form"
        # the developer usually doesn't need to know the public system assigned
        # name of the callback function.. we'll need it though to fake the request
        public_callback_name = func_name_by_form_id(form_id)
        request_args_json = '["%s"]' % form_id
        post = {}
        post[cls.PARAM_REQUEST] = public_callback_name
        post[cls.PARAM_ARGS] = request_args_json
        post["form_key"] = "value"
        post["form_key2"] = 15
        inst.set_data(post)

        files = {"file": "object", "passed": "here"}
        register_upload_callback(inst, form_id, my_callback, args_extra=(files,))
        process_request(inst)


        # let's see if we can override the default args and the response
        # class during callback registration
        params_custom = {cls.PARAM_RESPONSE_CLASS: CustomResponse}
        register_upload_callback(inst, form_id, my_callback2, args_extra=("custom1", "custom2"), **params_custom)
        process_request(inst)

        call_history_expected = [
            "my_callback", form_id, "UploadResponse", "object",
            "custom_callback", form_id, "CustomResponse", "custom1", "custom2"
        ]
        self.assertEqual(call_history_expected, call_history)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(SijaxMainTestCase))
    suite.addTest(unittest.makeSuite(SijaxStreamingTestCase))
    suite.addTest(unittest.makeSuite(SijaxCometTestCase))
    suite.addTest(unittest.makeSuite(SijaxUploadTestCase))
    
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')




# -*- coding: utf-8 -*-

"""
    sijax.helper
    ~~~~~~~~~~~~

    Provides various helper functions and objects that other modules use.

    :copyright: (c) 2011 by Slavi Pantaleev.
    :license: BSD, see LICENSE.txt for more details.
"""


from .exception import SijaxError


# Try to load the best json implementation,
# If json support is not available, we'll add
# an object that raises a RuntimeError when used
json = None
try:
    import simplejson as json
except ImportError:
    try:
        import json
    except ImportError:
        try:
            # Google Appengine offers simplejson via django
            from django.utils import simplejson as json
        except ImportError:
            pass
if json is None:
    class _JSON(object):
        def __getattr__(self, name):
            raise RuntimeError('You need a JSON library to use Sijax!')
    json = _JSON()


def init_static_path(static_path):
    """Mirrors the important static files from the whole Sijax package
    into a directory of your choice.

    It may be a good idea to run this whenever Sijax gets upgraded,
    so that your files will be kept in sync.

    The directory that you provide needs to be empty (if it exists),
    or to have been previously used by this same function.
    If the provided directory contains some other files, Sijax will refuse
    to use it, by raising a :class:`sijax.exception.SijaxError` exception.

    The following files will be made available in the specified directory:

    * ``sijax.js`` - the core javascript file used by Sijax
    * ``json2.js`` - JSON library that can be loaded for browsers
                     that don't support native JSON (like IE <= 7)
    * ``sijax_comet.js`` - the javascript file used by the Comet plugin
    * ``sijax_upload.js`` - the javascript file used by the Upload plugin
    * ``sijax_version`` - a system file that keeps track of versioning -
                          do not touch this file
    """

    import os, shutil, errno
    import sijax

    def mkdir_p(path):
        """Creates the whole directory tree (recursively), if it's missing."""
        try:
            os.makedirs(path)
        except OSError, exc:
            if exc.errno != errno.EEXIST:
                raise exc

    mkdir_p(static_path)

    version_file = os.path.join(static_path, 'sijax_version')

    if os.path.exists(version_file):
        version = open(version_file).read()
    else:
        version = None
        files_count = len(os.walk(static_path).next()[2])
        if files_count != 0:
            # non-empty path with a missing version file
            # this looks like a user directory - we'd better not touch anything!
            raise SijaxError('%s already contains files - refusing to write there!' %
                             static_path)

    if version == sijax.__version__:
        return

    if version is not None:
        # Cleanup previous files first
        for root, dirs, files in os.walk(static_path):
            for file_name in files:
                os.unlink(os.path.join(root, file_name))

    files = []

    core_js = os.path.join(os.path.dirname(sijax.__file__), 'js/')
    files.append(os.path.join(core_js, 'sijax.js'))
    files.append(os.path.join(core_js, 'json2.js'))

    plugins = os.path.join(os.path.dirname(sijax.__file__), 'plugin')
    files.append(os.path.join(plugins, 'comet/js/sijax_comet.js'))
    files.append(os.path.join(plugins, 'upload/js/sijax_upload.js'))

    for src_path in files:
        file_name = os.path.basename(src_path)
        dst_path = os.path.join(static_path, file_name)

        if not os.path.lexists(dst_path):
            shutil.copyfile(src_path, dst_path)

    fp = open(version_file, 'w')
    fp.write(sijax.__version__)
    fp.close()

